@@ -14,6 +14,9 @@
 #include <sys/stat.h>
 
 #ifdef _WIN32
+#ifndef UNICODE
+#define UNICODE
+#endif
 #include <windows.h>
 #include <ws2ipdef.h>
 #else
@@ -27,8 +30,6 @@
 
 #ifdef MMDB_DEBUG
 #define LOCAL
-#define NO_PROTO
-#define DEBUG_FUNC
 #define DEBUG_MSG(msg) fprintf(stderr, msg "\n")
 #define DEBUG_MSGF(fmt, ...) fprintf(stderr, fmt "\n", __VA_ARGS__)
 #define DEBUG_BINARY(fmt, byte)                                 \
@@ -44,7 +45,6 @@
 #define DEBUG_NL fprintf(stderr, "\n")
 #else
 #define LOCAL static
-#define NO_PROTO static
 #define DEBUG_MSG(...)
 #define DEBUG_MSGF(...)
 #define DEBUG_BINARY(...)
@@ -52,7 +52,7 @@
 #endif
 
 #ifdef MMDB_DEBUG
-DEBUG_FUNC char *byte_to_binary(uint8_t byte)
+char *byte_to_binary(uint8_t byte)
 {
     char *bits = malloc(sizeof(char) * 9);
     if (NULL == bits) {
@@ -67,7 +67,7 @@ DEBUG_FUNC char *byte_to_binary(uint8_t byte)
     return bits;
 }
 
-DEBUG_FUNC char *type_num_to_name(uint8_t num)
+char *type_num_to_name(uint8_t num)
 {
     switch (num) {
     case 0:
@@ -134,13 +134,11 @@ typedef struct record_info_s {
 // 64 leads us to allocating 4 KiB on a 64bit system.
 #define MMDB_POOL_INIT_SIZE 64
 
-/* *INDENT-OFF* */
-/* --prototypes automatically generated by dev-bin/regen-prototypes.pl - don't remove this comment */
 LOCAL int map_file(MMDB_s *const mmdb);
 LOCAL const uint8_t *find_metadata(const uint8_t *file_content,
                                    ssize_t file_size, uint32_t *metadata_size);
 LOCAL int read_metadata(MMDB_s *mmdb);
-LOCAL MMDB_s make_fake_metadata_db(MMDB_s *mmdb);
+LOCAL MMDB_s make_fake_metadata_db(const MMDB_s *const mmdb);
 LOCAL int value_for_key_as_uint16(MMDB_entry_s *start, char *key,
                                   uint16_t *value);
 LOCAL int value_for_key_as_uint32(MMDB_entry_s *start, char *key,
@@ -154,35 +152,34 @@ LOCAL int populate_languages_metadata(MMDB_s *mmdb, MMDB_s *metadata_db,
 LOCAL int populate_description_metadata(MMDB_s *mmdb, MMDB_s *metadata_db,
                                         MMDB_entry_s *metadata_start);
 LOCAL int resolve_any_address(const char *ipstr, struct addrinfo **addresses);
-LOCAL int find_address_in_search_tree(MMDB_s *mmdb, uint8_t *address,
+LOCAL int find_address_in_search_tree(const MMDB_s *const mmdb,
+                                      uint8_t *address,
                                       sa_family_t address_family,
                                       MMDB_lookup_result_s *result);
-LOCAL record_info_s record_info_for_database(MMDB_s *mmdb);
-LOCAL int find_ipv4_start_node(MMDB_s *mmdb);
-LOCAL uint8_t maybe_populate_result(MMDB_s *mmdb, uint32_t record,
-                                    uint16_t netmask,
-                                    MMDB_lookup_result_s *result);
-LOCAL uint8_t record_type(MMDB_s *const mmdb, uint64_t record);
+LOCAL record_info_s record_info_for_database(const MMDB_s *const mmdb);
+LOCAL int find_ipv4_start_node(MMDB_s *const mmdb);
+LOCAL uint8_t record_type(const MMDB_s *const mmdb, uint64_t record);
 LOCAL uint32_t get_left_28_bit_record(const uint8_t *record);
 LOCAL uint32_t get_right_28_bit_record(const uint8_t *record);
-LOCAL uint32_t data_section_offset_for_record(MMDB_s *const mmdb,
+LOCAL uint32_t data_section_offset_for_record(const MMDB_s *const mmdb,
                                               uint64_t record);
 LOCAL int path_length(va_list va_path);
-LOCAL int lookup_path_in_array(const char *path_elem, MMDB_s *mmdb,
+LOCAL int lookup_path_in_array(const char *path_elem, const MMDB_s *const mmdb,
                                MMDB_entry_data_s *entry_data);
-LOCAL int lookup_path_in_map(const char *path_elem, MMDB_s *mmdb,
+LOCAL int lookup_path_in_map(const char *path_elem, const MMDB_s *const mmdb,
                              MMDB_entry_data_s *entry_data);
-LOCAL int skip_map_or_array(MMDB_s *mmdb, MMDB_entry_data_s *entry_data);
-LOCAL int decode_one_follow(MMDB_s *mmdb, uint32_t offset,
+LOCAL int skip_map_or_array(const MMDB_s *const mmdb,
                             MMDB_entry_data_s *entry_data);
-LOCAL int decode_one(MMDB_s *mmdb, uint32_t offset,
+LOCAL int decode_one_follow(const MMDB_s *const mmdb, uint32_t offset,
+                            MMDB_entry_data_s *entry_data);
+LOCAL int decode_one(const MMDB_s *const mmdb, uint32_t offset,
                      MMDB_entry_data_s *entry_data);
 LOCAL int get_ext_type(int raw_ext_type);
 LOCAL uint32_t get_ptr_from(uint8_t ctrl, uint8_t const *const ptr,
                             int ptr_size);
-LOCAL int get_entry_data_list(MMDB_s *mmdb,
+LOCAL int get_entry_data_list(const MMDB_s *const mmdb,
                               uint32_t offset,
-                              size_t const index,
+                              MMDB_entry_data_list_s *const entry_data_list,
                               MMDB_data_pool_s *const pool,
                               int depth);
 LOCAL float get_ieee754_float(const uint8_t *restrict p);
@@ -200,8 +197,6 @@ LOCAL MMDB_entry_data_list_s *dump_entry_data_list(
     int *status);
 LOCAL void print_indentation(FILE *stream, int i);
 LOCAL char *bytes_to_hex(uint8_t *bytes, uint32_t size);
-/* --prototypes end - don't remove this comment-- */
-/* *INDENT-ON* */
 
 #define CHECKED_DECODE_ONE(mmdb, offset, entry_data)                        \
     do {                                                                    \
@@ -233,6 +228,7 @@ int MMDB_open(const char *const filename, uint32_t flags, MMDB_s *const mmdb)
     mmdb->data_section = NULL;
     mmdb->metadata.database_type = NULL;
     mmdb->metadata.languages.count = 0;
+    mmdb->metadata.languages.names = NULL;
     mmdb->metadata.description.count = 0;
 
     mmdb->filename = mmdb_strdup(filename);
@@ -322,13 +318,33 @@ int MMDB_open(const char *const filename, uint32_t flags, MMDB_s *const mmdb)
 
 #ifdef _WIN32
 
+LOCAL LPWSTR utf8_to_utf16(const char *utf8_str)
+{
+    int wide_chars = MultiByteToWideChar(CP_UTF8, 0, utf8_str, -1, NULL, 0);
+    wchar_t *utf16_str = (wchar_t *)malloc(wide_chars * sizeof(wchar_t));
+
+    if (MultiByteToWideChar(CP_UTF8, 0, utf8_str, -1, utf16_str,
+                            wide_chars) < 1) {
+        free(utf16_str);
+        return NULL;
+    }
+
+    return utf16_str;
+}
+
 LOCAL int map_file(MMDB_s *const mmdb)
 {
     DWORD size;
     int status = MMDB_SUCCESS;
     HANDLE mmh = NULL;
-    HANDLE fd = CreateFileA(mmdb->filename, GENERIC_READ, FILE_SHARE_READ, NULL,
-                            OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
+    HANDLE fd = INVALID_HANDLE_VALUE;
+    LPWSTR utf16_filename = utf8_to_utf16(mmdb->filename);
+    if (!utf16_filename) {
+        status = MMDB_FILE_OPEN_ERROR;
+        goto cleanup;
+    }
+    fd = CreateFile(utf16_filename, GENERIC_READ, FILE_SHARE_READ, NULL,
+                    OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
     if (fd == INVALID_HANDLE_VALUE) {
         status = MMDB_FILE_OPEN_ERROR;
         goto cleanup;
@@ -338,7 +354,7 @@ LOCAL int map_file(MMDB_s *const mmdb)
         status = MMDB_FILE_OPEN_ERROR;
         goto cleanup;
     }
-    mmh = CreateFileMappingA(fd, NULL, PAGE_READONLY, 0, size, NULL);
+    mmh = CreateFileMapping(fd, NULL, PAGE_READONLY, 0, size, NULL);
     /* Microsoft documentation for CreateFileMapping indicates this returns
         NULL not INVALID_HANDLE_VALUE on error */
     if (NULL == mmh) {
@@ -364,18 +380,23 @@ LOCAL int map_file(MMDB_s *const mmdb)
         CloseHandle(mmh);
     }
     errno = saved_errno;
+    free(utf16_filename);
 
     return status;
 }
 
-#else
+#else // _WIN32
 
 LOCAL int map_file(MMDB_s *const mmdb)
 {
     ssize_t size;
     int status = MMDB_SUCCESS;
 
-    int fd = open(mmdb->filename, O_RDONLY);
+    int flags = O_RDONLY;
+#ifdef O_CLOEXEC
+    flags |= O_CLOEXEC;
+#endif
+    int fd = open(mmdb->filename, flags);
     struct stat s;
     if (fd < 0 || fstat(fd, &s)) {
         status = MMDB_FILE_OPEN_ERROR;
@@ -412,7 +433,7 @@ LOCAL int map_file(MMDB_s *const mmdb)
     return status;
 }
 
-#endif
+#endif // _WIN32
 
 LOCAL const uint8_t *find_metadata(const uint8_t *file_content,
                                    ssize_t file_size, uint32_t *metadata_size)
@@ -562,7 +583,7 @@ LOCAL int read_metadata(MMDB_s *mmdb)
     return MMDB_SUCCESS;
 }
 
-LOCAL MMDB_s make_fake_metadata_db(MMDB_s *mmdb)
+LOCAL MMDB_s make_fake_metadata_db(const MMDB_s *const mmdb)
 {
     MMDB_s fake_metadata_db = {
         .data_section      = mmdb->metadata_section,
@@ -810,7 +831,7 @@ LOCAL int populate_description_metadata(MMDB_s *mmdb, MMDB_s *metadata_db,
     return status;
 }
 
-MMDB_lookup_result_s MMDB_lookup_string(MMDB_s *const mmdb,
+MMDB_lookup_result_s MMDB_lookup_string(const MMDB_s *const mmdb,
                                         const char *const ipstr,
                                         int *const gai_error,
                                         int *const mmdb_error)
@@ -856,7 +877,7 @@ LOCAL int resolve_any_address(const char *ipstr, struct addrinfo **addresses)
 }
 
 MMDB_lookup_result_s MMDB_lookup_sockaddr(
-    MMDB_s *const mmdb,
+    const MMDB_s *const mmdb,
     const struct sockaddr *const sockaddr,
     int *const mmdb_error)
 {
@@ -896,7 +917,8 @@ MMDB_lookup_result_s MMDB_lookup_sockaddr(
     return result;
 }
 
-LOCAL int find_address_in_search_tree(MMDB_s *mmdb, uint8_t *address,
+LOCAL int find_address_in_search_tree(const MMDB_s *const mmdb,
+                                      uint8_t *address,
                                       sa_family_t address_family,
                                       MMDB_lookup_result_s *result)
 {
@@ -905,82 +927,51 @@ LOCAL int find_address_in_search_tree(MMDB_s *mmdb, uint8_t *address,
         return MMDB_UNKNOWN_DATABASE_FORMAT_ERROR;
     }
 
-    DEBUG_NL;
-    DEBUG_MSG("Looking for address in search tree");
-
     uint32_t value = 0;
-    uint16_t max_depth0 = mmdb->depth - 1;
-    uint16_t start_bit = max_depth0;
-
+    uint16_t current_bit = 0;
     if (mmdb->metadata.ip_version == 6 && address_family == AF_INET) {
-        int mmdb_error = find_ipv4_start_node(mmdb);
-        if (MMDB_SUCCESS != mmdb_error) {
-            return mmdb_error;
-        }
-        DEBUG_MSGF("IPv4 start node is %u (netmask %u)",
-                   mmdb->ipv4_start_node.node_value,
-                   mmdb->ipv4_start_node.netmask);
-
-        uint8_t type = maybe_populate_result(mmdb,
-                                             mmdb->ipv4_start_node.node_value,
-                                             mmdb->ipv4_start_node.netmask,
-                                             result);
-        if (MMDB_RECORD_TYPE_INVALID == type) {
-            return MMDB_CORRUPT_SEARCH_TREE_ERROR;
-        }
-
-        /* We have an IPv6 database with no IPv4 data */
-        if (MMDB_RECORD_TYPE_SEARCH_NODE != type) {
-            return MMDB_SUCCESS;
-        }
-
         value = mmdb->ipv4_start_node.node_value;
-        start_bit -= mmdb->ipv4_start_node.netmask;
+        current_bit = mmdb->ipv4_start_node.netmask;
     }
 
+    uint32_t node_count = mmdb->metadata.node_count;
     const uint8_t *search_tree = mmdb->file_content;
     const uint8_t *record_pointer;
-    for (int current_bit = start_bit; current_bit >= 0; current_bit--) {
-        uint8_t bit_is_true =
-            address[(max_depth0 - current_bit) >> 3]
-            & (1U << (~(max_depth0 - current_bit) & 7)) ? 1 : 0;
-
-        DEBUG_MSGF("Looking at bit %i - bit's value is %i", current_bit,
-                   bit_is_true);
-        DEBUG_MSGF("  current node = %u", value);
+    for (; current_bit < mmdb->depth && value < node_count; current_bit++) {
+        uint8_t bit = 1U &
+                      (address[current_bit >> 3] >> (7 - (current_bit % 8)));
 
         record_pointer = &search_tree[value * record_info.record_length];
         if (record_pointer + record_info.record_length > mmdb->data_section) {
             return MMDB_CORRUPT_SEARCH_TREE_ERROR;
         }
-        if (bit_is_true) {
+        if (bit) {
             record_pointer += record_info.right_record_offset;
             value = record_info.right_record_getter(record_pointer);
         } else {
             value = record_info.left_record_getter(record_pointer);
         }
-
-        uint8_t type = maybe_populate_result(mmdb, value, (uint16_t)current_bit,
-                                             result);
-        if (MMDB_RECORD_TYPE_INVALID == type) {
-            return MMDB_CORRUPT_SEARCH_TREE_ERROR;
-        }
-
-        if (MMDB_RECORD_TYPE_SEARCH_NODE != type) {
-            return MMDB_SUCCESS;
-        }
-
-        DEBUG_MSGF("  proceeding to search tree node %i", value);
     }
 
-    DEBUG_MSG(
-        "Reached the end of the address bits without leaving the search tree");
+    result->netmask = current_bit;
 
-    // We should not be able to reach this return. If we do, something very bad happened.
-    return MMDB_CORRUPT_SEARCH_TREE_ERROR;
+    if (value >= node_count + mmdb->data_section_size) {
+        // The pointer points off the end of the database.
+        return MMDB_CORRUPT_SEARCH_TREE_ERROR;
+    }
+
+    if (value == node_count) {
+        // record is empty
+        result->found_entry = false;
+        return MMDB_SUCCESS;
+    }
+    result->found_entry = true;
+    result->entry.offset = data_section_offset_for_record(mmdb, value);
+
+    return MMDB_SUCCESS;
 }
 
-LOCAL record_info_s record_info_for_database(MMDB_s *mmdb)
+LOCAL record_info_s record_info_for_database(const MMDB_s *const mmdb)
 {
     record_info_s record_info = {
         .record_length       = mmdb->full_record_byte_size,
@@ -1006,7 +997,7 @@ LOCAL record_info_s record_info_for_database(MMDB_s *mmdb)
     return record_info;
 }
 
-LOCAL int find_ipv4_start_node(MMDB_s *mmdb)
+LOCAL int find_ipv4_start_node(MMDB_s *const mmdb)
 {
     /* In a pathological case of a database with a single node search tree,
      * this check will be true even after we've found the IPv4 start node, but
@@ -1021,17 +1012,14 @@ LOCAL int find_ipv4_start_node(MMDB_s *mmdb)
     uint32_t node_value = 0;
     const uint8_t *record_pointer;
     uint16_t netmask;
-    for (netmask = 0; netmask < 96; netmask++) {
+    uint32_t node_count = mmdb->metadata.node_count;
+
+    for (netmask = 0; netmask < 96 && node_value < node_count; netmask++) {
         record_pointer = &search_tree[node_value * record_info.record_length];
         if (record_pointer + record_info.record_length > mmdb->data_section) {
             return MMDB_CORRUPT_SEARCH_TREE_ERROR;
         }
         node_value = record_info.left_record_getter(record_pointer);
-        /* This can happen if there's no IPv4 data _or_ if there is a subnet
-         * with data that contains the entire IPv4 range (like ::/64) */
-        if (node_value >= mmdb->metadata.node_count) {
-            break;
-        }
     }
 
     mmdb->ipv4_start_node.node_value = node_value;
@@ -1040,29 +1028,7 @@ LOCAL int find_ipv4_start_node(MMDB_s *mmdb)
     return MMDB_SUCCESS;
 }
 
-LOCAL uint8_t maybe_populate_result(MMDB_s *mmdb, uint32_t record,
-                                    uint16_t netmask,
-                                    MMDB_lookup_result_s *result)
-{
-    uint8_t type = record_type(mmdb, record);
-
-    if (MMDB_RECORD_TYPE_SEARCH_NODE == type ||
-        MMDB_RECORD_TYPE_INVALID == type) {
-        return type;
-    }
-
-    result->netmask = mmdb->depth - netmask;
-
-    result->entry.offset = data_section_offset_for_record(mmdb, record);
-
-    // type is either MMDB_RECORD_TYPE_DATA or MMDB_RECORD_TYPE_EMPTY
-    // at this point
-    result->found_entry = MMDB_RECORD_TYPE_DATA == type;
-
-    return type;
-}
-
-LOCAL uint8_t record_type(MMDB_s *const mmdb, uint64_t record)
+LOCAL uint8_t record_type(const MMDB_s *const mmdb, uint64_t record)
 {
     uint32_t node_count = mmdb->metadata.node_count;
 
@@ -1102,7 +1068,7 @@ LOCAL uint32_t get_right_28_bit_record(const uint8_t *record)
     return value & 0xfffffff;
 }
 
-int MMDB_read_node(MMDB_s *const mmdb, uint32_t node_number,
+int MMDB_read_node(const MMDB_s *const mmdb, uint32_t node_number,
                    MMDB_search_node_s *const node)
 {
     record_info_s record_info = record_info_for_database(mmdb);
@@ -1139,7 +1105,7 @@ int MMDB_read_node(MMDB_s *const mmdb, uint32_t node_number,
     return MMDB_SUCCESS;
 }
 
-LOCAL uint32_t data_section_offset_for_record(MMDB_s *const mmdb,
+LOCAL uint32_t data_section_offset_for_record(const MMDB_s *const mmdb,
                                               uint64_t record)
 {
     return (uint32_t)record - mmdb->metadata.node_count -
@@ -1206,7 +1172,7 @@ int MMDB_aget_value(MMDB_entry_s *const start,
                     MMDB_entry_data_s *const entry_data,
                     const char *const *const path)
 {
-    MMDB_s *mmdb = start->mmdb;
+    const MMDB_s *const mmdb = start->mmdb;
     uint32_t offset = start->offset;
 
     memset(entry_data, 0, sizeof(MMDB_entry_data_s));
@@ -1259,7 +1225,8 @@ int MMDB_aget_value(MMDB_entry_s *const start,
     return MMDB_SUCCESS;
 }
 
-LOCAL int lookup_path_in_array(const char *path_elem, MMDB_s *mmdb,
+LOCAL int lookup_path_in_array(const char *path_elem,
+                               const MMDB_s *const mmdb,
                                MMDB_entry_data_s *entry_data)
 {
     uint32_t size = entry_data->data_size;
@@ -1268,12 +1235,20 @@ LOCAL int lookup_path_in_array(const char *path_elem, MMDB_s *mmdb,
     int saved_errno = errno;
     errno = 0;
     int array_index = strtol(path_elem, &first_invalid, 10);
-    if (array_index < 0 || ERANGE == errno) {
+    if (ERANGE == errno) {
         errno = saved_errno;
         return MMDB_INVALID_LOOKUP_PATH_ERROR;
     }
     errno = saved_errno;
 
+    if (array_index < 0) {
+        array_index += size;
+
+        if (array_index < 0) {
+            return MMDB_LOOKUP_PATH_DOES_NOT_MATCH_DATA_ERROR;
+        }
+    }
+
     if (*first_invalid || (uint32_t)array_index >= size) {
         return MMDB_LOOKUP_PATH_DOES_NOT_MATCH_DATA_ERROR;
     }
@@ -1295,7 +1270,8 @@ LOCAL int lookup_path_in_array(const char *path_elem, MMDB_s *mmdb,
     return MMDB_SUCCESS;
 }
 
-LOCAL int lookup_path_in_map(const char *path_elem, MMDB_s *mmdb,
+LOCAL int lookup_path_in_map(const char *path_elem,
+                             const MMDB_s *const mmdb,
                              MMDB_entry_data_s *entry_data)
 {
     uint32_t size = entry_data->data_size;
@@ -1336,7 +1312,8 @@ LOCAL int lookup_path_in_map(const char *path_elem, MMDB_s *mmdb,
     return MMDB_LOOKUP_PATH_DOES_NOT_MATCH_DATA_ERROR;
 }
 
-LOCAL int skip_map_or_array(MMDB_s *mmdb, MMDB_entry_data_s *entry_data)
+LOCAL int skip_map_or_array(const MMDB_s *const mmdb,
+                            MMDB_entry_data_s *entry_data)
 {
     if (entry_data->type == MMDB_DATA_TYPE_MAP) {
         uint32_t size = entry_data->data_size;
@@ -1362,7 +1339,7 @@ LOCAL int skip_map_or_array(MMDB_s *mmdb, MMDB_entry_data_s *entry_data)
     return MMDB_SUCCESS;
 }
 
-LOCAL int decode_one_follow(MMDB_s *mmdb, uint32_t offset,
+LOCAL int decode_one_follow(const MMDB_s *const mmdb, uint32_t offset,
                             MMDB_entry_data_s *entry_data)
 {
     CHECKED_DECODE_ONE(mmdb, offset, entry_data);
@@ -1392,7 +1369,7 @@ LOCAL int decode_one_follow(MMDB_s *mmdb, uint32_t offset,
 }
 
 #if !MMDB_UINT128_IS_BYTE_ARRAY
-NO_PROTO mmdb_uint128_t get_uint128(const uint8_t *p, int length)
+LOCAL mmdb_uint128_t get_uint128(const uint8_t *p, int length)
 {
     mmdb_uint128_t value = 0;
     while (length-- > 0) {
@@ -1403,7 +1380,7 @@ NO_PROTO mmdb_uint128_t get_uint128(const uint8_t *p, int length)
 }
 #endif
 
-LOCAL int decode_one(MMDB_s *mmdb, uint32_t offset,
+LOCAL int decode_one(const MMDB_s *const mmdb, uint32_t offset,
                      MMDB_entry_data_s *entry_data)
 {
     const uint8_t *mem = mmdb->data_section;
@@ -1633,7 +1610,7 @@ LOCAL uint32_t get_ptr_from(uint8_t ctrl, uint8_t const *const ptr,
 }
 
 int MMDB_get_metadata_as_entry_data_list(
-    MMDB_s *const mmdb, MMDB_entry_data_list_s **const entry_data_list)
+    const MMDB_s *const mmdb, MMDB_entry_data_list_s **const entry_data_list)
 {
     MMDB_s metadata_db = make_fake_metadata_db(mmdb);
 
@@ -1653,32 +1630,27 @@ int MMDB_get_entry_data_list(
         return MMDB_OUT_OF_MEMORY_ERROR;
     }
 
-    size_t index = 0;
-    if (data_pool_alloc(pool, &index) != 0) {
+    MMDB_entry_data_list_s *const list = data_pool_alloc(pool);
+    if (!list) {
         data_pool_destroy(pool);
         return MMDB_OUT_OF_MEMORY_ERROR;
     }
 
-    int const status = get_entry_data_list(start->mmdb, start->offset, index,
+    int const status = get_entry_data_list(start->mmdb, start->offset, list,
                                            pool, 0);
 
     *entry_data_list = data_pool_to_list(pool);
     if (!*entry_data_list) {
         data_pool_destroy(pool);
-        // This error code is not particularly accurate. However this really
-        // shouldn't happen.
         return MMDB_OUT_OF_MEMORY_ERROR;
     }
 
-    // The caller doesn't know about the pool. Just the list.
-    free(pool);
-
     return status;
 }
 
-LOCAL int get_entry_data_list(MMDB_s *mmdb,
+LOCAL int get_entry_data_list(const MMDB_s *const mmdb,
                               uint32_t offset,
-                              size_t const index,
+                              MMDB_entry_data_list_s *const entry_data_list,
                               MMDB_data_pool_s *const pool,
                               int depth)
 {
@@ -1687,17 +1659,6 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb,
         return MMDB_INVALID_DATA_ERROR;
     }
     depth++;
-
-    // Note it is important to never re-use memory pointing into the pool after
-    // it may have changed in size. This is because the memory location may
-    // move when it grows. In practice this means that if we ever call
-    // data_pool_alloc() or this function recursively, all MMDB_entry_data_list
-    // pointers should be considered invalid, and you should look them up again
-    // by their index.
-    MMDB_entry_data_list_s *entry_data_list = data_pool_lookup(pool, index);
-    if (!entry_data_list) {
-        return MMDB_INVALID_DATA_ERROR;
-    }
     CHECKED_DECODE_ONE(mmdb, offset, &entry_data_list->entry_data);
 
     switch (entry_data_list->entry_data.type) {
@@ -1718,18 +1679,14 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb,
             if (entry_data_list->entry_data.type == MMDB_DATA_TYPE_ARRAY
                 || entry_data_list->entry_data.type == MMDB_DATA_TYPE_MAP) {
 
-                int status = get_entry_data_list(mmdb, last_offset, index, pool,
-                                                 depth);
+                int status =
+                    get_entry_data_list(mmdb, last_offset, entry_data_list,
+                                        pool, depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on pointer failed.");
                     return status;
                 }
             }
-
-            entry_data_list = data_pool_lookup(pool, index);
-            if (!entry_data_list) {
-                return MMDB_INVALID_DATA_ERROR;
-            }
             entry_data_list->entry_data.offset_to_next = next_offset;
         }
         break;
@@ -1738,31 +1695,24 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb,
             uint32_t array_size = entry_data_list->entry_data.data_size;
             uint32_t array_offset = entry_data_list->entry_data.offset_to_next;
             while (array_size-- > 0) {
-                size_t index2 = 0;
-                if (data_pool_alloc(pool, &index2) != 0) {
+                MMDB_entry_data_list_s *entry_data_list_to =
+                    data_pool_alloc(pool);
+                if (!entry_data_list_to) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                int status = get_entry_data_list(mmdb, array_offset, index2,
-                                                 pool, depth);
+                int status =
+                    get_entry_data_list(mmdb, array_offset, entry_data_list_to,
+                                        pool, depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on array element failed.");
                     return status;
                 }
 
-                MMDB_entry_data_list_s *entry_data_list_to
-                    = data_pool_lookup(pool, index2);
-                if (!entry_data_list_to) {
-                    return MMDB_INVALID_DATA_ERROR;
-                }
                 array_offset = entry_data_list_to->entry_data.offset_to_next;
             }
-
-            entry_data_list = data_pool_lookup(pool, index);
-            if (!entry_data_list) {
-                return MMDB_INVALID_DATA_ERROR;
-            }
             entry_data_list->entry_data.offset_to_next = array_offset;
+
         }
         break;
     case MMDB_DATA_TYPE_MAP:
@@ -1771,47 +1721,32 @@ LOCAL int get_entry_data_list(MMDB_s *mmdb,
 
             offset = entry_data_list->entry_data.offset_to_next;
             while (size-- > 0) {
-                size_t key_index = 0;
-                if (data_pool_alloc(pool, &key_index) != 0) {
+                MMDB_entry_data_list_s *list_key = data_pool_alloc(pool);
+                if (!list_key) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                int status = get_entry_data_list(mmdb, offset, key_index, pool,
-                                                 depth);
+                int status =
+                    get_entry_data_list(mmdb, offset, list_key, pool, depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on map key failed.");
                     return status;
                 }
 
-                MMDB_entry_data_list_s *entry_data_list_to
-                    = data_pool_lookup(pool, key_index);
-                if (!entry_data_list_to) {
-                    return MMDB_INVALID_DATA_ERROR;
-                }
-                offset = entry_data_list_to->entry_data.offset_to_next;
+                offset = list_key->entry_data.offset_to_next;
 
-                size_t value_index = 0;
-                if (data_pool_alloc(pool, &value_index) != 0) {
+                MMDB_entry_data_list_s *list_value = data_pool_alloc(pool);
+                if (!list_value) {
                     return MMDB_OUT_OF_MEMORY_ERROR;
                 }
 
-                status = get_entry_data_list(mmdb, offset, value_index, pool,
+                status = get_entry_data_list(mmdb, offset, list_value, pool,
                                              depth);
                 if (MMDB_SUCCESS != status) {
                     DEBUG_MSG("get_entry_data_list on map element failed.");
                     return status;
                 }
-
-                entry_data_list_to = data_pool_lookup(pool, value_index);
-                if (!entry_data_list_to) {
-                    return MMDB_INVALID_DATA_ERROR;
-                }
-                offset = entry_data_list_to->entry_data.offset_to_next;
-            }
-
-            entry_data_list = data_pool_lookup(pool, index);
-            if (!entry_data_list) {
-                return MMDB_INVALID_DATA_ERROR;
+                offset = list_value->entry_data.offset_to_next;
             }
             entry_data_list->entry_data.offset_to_next = offset;
         }
@@ -1895,7 +1830,7 @@ void MMDB_free_entry_data_list(MMDB_entry_data_list_s *const entry_data_list)
     if (entry_data_list == NULL) {
         return;
     }
-    free(entry_data_list);
+    data_pool_destroy(entry_data_list->pool);
 }
 
 void MMDB_close(MMDB_s *const mmdb)
@@ -1933,7 +1868,7 @@ LOCAL void free_mmdb_struct(MMDB_s *const mmdb)
 
 LOCAL void free_languages_metadata(MMDB_s *mmdb)
 {
-    if (!mmdb->metadata.languages.count) {
+    if (!mmdb->metadata.languages.names) {
         return;
     }
 
