@@ -11,7 +11,7 @@ void test_all_data_types(MMDB_lookup_result_s *result, const char *ip,
         MMDB_entry_data_s data =
             data_ok(result, MMDB_DATA_TYPE_UTF8_STRING, description,
                     "utf8_string", NULL);
-        const char *string = strndup(data.utf8_string, data.data_size);
+        const char *string = mmdb_strndup(data.utf8_string, data.data_size);
         // This is hex for "unicode! ☯ - ♫" as bytes
         char expect[19] =
         { 0x75, 0x6e, 0x69, 0x63, 0x6f, 0x64, 0x65, 0x21, 0x20, 0xe2, 0x98,
@@ -69,7 +69,7 @@ void test_all_data_types(MMDB_lookup_result_s *result, const char *ip,
         MMDB_entry_data_s data =
             data_ok(result, MMDB_DATA_TYPE_UINT32, description, "uint32", NULL);
         uint32_t expect = 1 << 28;
-        ok(data.uint32 == expect, "uint32 field is 2**28");
+        cmp_ok(data.uint32, "==", expect, "uint32 field is 2**28");
     }
 
     {
@@ -91,7 +91,7 @@ void test_all_data_types(MMDB_lookup_result_s *result, const char *ip,
             data_ok(result, MMDB_DATA_TYPE_UINT64, description, "uint64", NULL);
         uint64_t expect = 1;
         expect <<= 60;
-        ok(data.uint64 == expect, "uint64 field is 2**60");
+        cmp_ok(data.uint64, "==", expect, "uint64 field is 2**60");
     }
 
     {
@@ -108,7 +108,7 @@ void test_all_data_types(MMDB_lookup_result_s *result, const char *ip,
 #else
         mmdb_uint128_t expect = 1;
         expect <<= 120;
-        ok(data.uint128 == expect, "uint128 field is 2**120");
+        cmp_ok(data.uint128, "==", expect, "uint128 field is 2**120");
 #endif
     }
 
@@ -170,7 +170,7 @@ void test_all_data_types(MMDB_lookup_result_s *result, const char *ip,
         data =
             data_ok(result, MMDB_DATA_TYPE_UTF8_STRING, description, "map",
                     "mapX", "utf8_stringX", NULL);
-        const char *string = strndup(data.utf8_string, data.data_size);
+        const char *string = mmdb_strndup(data.utf8_string, data.data_size);
         is(string, "hello", "map{mapX}{utf8_stringX} is 'hello'");
         free((char *)string);
 
@@ -269,7 +269,7 @@ void test_all_data_types_as_zero(MMDB_lookup_result_s *result, const char *ip,
         MMDB_entry_data_s data =
             data_ok(result, MMDB_DATA_TYPE_UINT32, description, "uint32", NULL);
         uint32_t expect = 0;
-        ok(data.uint32 == expect, "uint32 field is 0");
+        cmp_ok(data.uint32, "==", expect, "uint32 field is 0");
     }
 
     {
@@ -290,7 +290,7 @@ void test_all_data_types_as_zero(MMDB_lookup_result_s *result, const char *ip,
         MMDB_entry_data_s data =
             data_ok(result, MMDB_DATA_TYPE_UINT64, description, "uint64", NULL);
         uint64_t expect = 0;
-        ok(data.uint64 == expect, "uint64 field is 0");
+        cmp_ok(data.uint64, "==", expect, "uint64 field is 0");
     }
 
     {
@@ -306,7 +306,7 @@ void test_all_data_types_as_zero(MMDB_lookup_result_s *result, const char *ip,
         ok(memcmp(data.uint128, expect, 16) == 0, "uint128 field is 0");
 #else
         mmdb_uint128_t expect = 0;
-        ok(data.uint128 == expect, "uint128 field is 0");
+        cmp_ok(data.uint128, "==", expect, "uint128 field is 0");
 #endif
     }
 
